@@ -12,7 +12,7 @@ import Network.Socket
 import Network.Socket.ByteString
 import Network.Test.Common
 import System.Mem (performGC)
-import System.IO.Error (tryIOError, isAlreadyInUseError)
+import System.IO.Error (tryIOError)
 import System.IO.Temp (withSystemTempDirectory)
 import Foreign.C.Types ()
 
@@ -70,9 +70,12 @@ spec = do
             bind sock (addrAddress addr) `shouldThrow` anyIOException
 #endif
 
-        it "successfully binds to a unix socket, twice" $ do
+        it "successfully binds to a unix socket" $ do
             withSystemTempDirectory "haskell-network" $ \path -> do
                 let sfile = path ++ "/socket-file"
+                -- exist <- doesFileExist sfile
+                -- when exist $ removeFile sfile
+                -- removeFile sfile
                 let addr = SockAddrUnix sfile
                 when (isSupportedSockAddr addr) $ do
                     sock0 <- socket AF_UNIX Stream defaultProtocol
@@ -82,23 +85,10 @@ spec = do
                     sock1 <- socket AF_UNIX Stream defaultProtocol
                     tryIOError (bind sock1 addr) >>= \o -> case o of
                         Right () -> error "bind should have failed but succeeded"
-                        Left e | not (isAlreadyInUseError e) -> ioError e
-                        _ -> return ()
+                        _        -> return ()
 
                     close sock0
-
-                    -- Unix systems tend to leave the file existing, which is
-                    -- why our `bind` does its workaround. however if any
-                    -- system in the future does fix this issue, we don't want
-                    -- this test to fail, since that would defeat the purpose
-                    -- of our workaround. but you can uncomment the below lines
-                    -- if you want to play with this on your own system.
-                    --import System.Directory (doesPathExist)
-                    --ex <- doesPathExist sfile
-                    --unless ex $ error "socket file was deleted unexpectedly"
-
-                    sock2 <- socket AF_UNIX Stream defaultProtocol
-                    bind sock2 addr
+                    close sock1
 
     describe "UserTimeout" $ do
         it "can be set" $ do
